//
//  Playfield.m
//  Demine
//
//  Created by Michael Heyeck on 4/9/10.
//  Copyright Fair Oaks Labs, Inc. 2010. All rights reserved.
//

#import <QuartzCore/QuartzCore.h>
#import "UIButton+NavButton.h"
#import "Path.h"
#import "Text.h"
#import "RootViewController.h"
#import "Playfield.h"


const static NSInteger TAG_BASE = 501;

const static CGSize tileSize = { 64,  64};
const static CGSize cellSize = { 64,  64};

UIKIT_EXTERN NSString* const UIApplicationDidEnterBackgroundNotification __attribute__((weak_import));


@interface Playfield ()

@property (nonatomic, retain, readonly) UIBarButtonItem* backButton;
@property (nonatomic, retain, readonly) UIBarButtonItem* retryButton;

@property (nonatomic, retain) NSTimer* daemonTimer;
@property (nonatomic, retain, readonly) NSDateFormatter* timerFormatter;
@property (nonatomic, retain, readonly) NSArray* box;
@property (nonatomic, retain, readonly) NSArray* glyphs;
@property (nonatomic, retain, readonly) NSArray* contents;
@property (nonatomic, retain) NSDate* timerStart;

- (void)pause;
- (void)unpause;
- (void)finishUnpause:(NSString*)animationID finished:(NSNumber*)finished context:(void*)context;
- (void)save;
- (void)retry;
- (void)dismiss;

- (BOOL)cellForPoint:(CGPoint*)point;
- (UIImage*)imageForCell:(CGPoint)pos;
- (void)setHeader;
- (void)startTimer;
- (void)stopTimer;

- (void)removeTiles;
- (void)createTiles;
- (UIView*)addTileForFrame:(CGRect)frame;
- (void)sizeContent;
- (void)reload;

@end


@implementation Playfield

@synthesize daemonTimer;
@synthesize timerStart;
@synthesize titleLabel;
@synthesize toolbar;
@synthesize	scrollView;
@synthesize pauseScreen;
@synthesize pauseLabel;
@synthesize modeControl;
@synthesize content;

- (UIBarButtonItem*)backButton
{
	if ((!backButton) && (self.navigationController.viewControllers.count >= 2))
	{
		backButton = [[UIBarButtonItem alloc] initWithCustomView:[[UIButton buttonWithType:UIButtonTypeCustom] configureForBackButtonWithTitle:@"Menu" target:self action:@selector(dismiss)]];
	}
	return backButton;
}

- (UIBarButtonItem*)retryButton
{
	if (!retryButton)
	{
		retryButton = [[UIBarButtonItem alloc] initWithTitle:@"Retry" style:UIBarButtonItemStyleBordered target:self action:@selector(retry)];
	}
	return retryButton;
}

- (NSDateFormatter*)timerFormatter
{
	if (!timerFormatter)
	{
		timerFormatter = [[NSDateFormatter alloc] init];
		timerFormatter.dateFormat = @"H:mm:ss";
		timerFormatter.timeZone = [NSTimeZone timeZoneForSecondsFromGMT:0];
	}
	return timerFormatter;
}

- (NSArray*)box
{
	if (!box)
	{
		// Basic (top) trapezoid
		CGFloat w = 5;
		CGPoint edge[4]  = {CGPointMake(0, 0), CGPointMake(cellSize.width, 0), CGPointMake(cellSize.width-w, w), CGPointMake(w, w)};
		CGPoint floor[4]  = {CGPointMake(w, w), CGPointMake(cellSize.width-w, w), CGPointMake(cellSize.width-w, cellSize.width-w), CGPointMake(w, cellSize.width-w)};

		// Paths
		CGMutablePathRef p;
		CGAffineTransform m;
		// Floor
		m = CGAffineTransformMakeScale(1.0, 1.0);
		Path* f = [[Path new] autorelease];
		p = CGPathCreateMutable();
		CGPathAddLines(p, &m, floor, 4);
		f.path = p;
		f.color = [UIColor grayColor];
		f.drawMode = kCGPathFill;
		CGPathRelease(p);
		// Top
		m = CGAffineTransformMakeScale(1.0, 1.0);
		Path* t = [[Path new] autorelease];
		p = CGPathCreateMutable();
		CGPathAddLines(p, &m, edge, 4);
		t.path = p;
		t.color = [UIColor colorWithWhite:.2 alpha:1.0];
		t.drawMode = kCGPathFill;
		CGPathRelease(p);
		// Bottom
		m = CGAffineTransformConcat(CGAffineTransformMakeScale(1.0, -1.0),
									CGAffineTransformMakeTranslation(0, cellSize.height));
		Path* b = [[Path new] autorelease];
		p = CGPathCreateMutable();
		CGPathAddLines(p, &m, edge, 4);
		b.path = p;
		b.color = [UIColor colorWithWhite:.7 alpha:1.0];
		b.drawMode = kCGPathFill;
		CGPathRelease(p);
		// Left
		m = CGAffineTransformConcat(CGAffineTransformMakeRotation(M_PI/2),
									CGAffineTransformMakeScale(-1.0, 1.0));
		Path* l = [[Path new] autorelease];
		p = CGPathCreateMutable();
		CGPathAddLines(p, &m, edge, 4);
		l.path = p;
		l.color = [UIColor colorWithWhite:.1 alpha:1.0];
		l.drawMode = kCGPathFill;
		CGPathRelease(p);
		// Right
		m = CGAffineTransformConcat(CGAffineTransformMakeRotation(M_PI/2),
									CGAffineTransformMakeTranslation(cellSize.width, 0));
		Path* r = [[Path new] autorelease];
		p = CGPathCreateMutable();
		CGPathAddLines(p, &m, edge, 4);
		r.path = p;
		r.color = [UIColor colorWithWhite:.6 alpha:1.0];
		r.drawMode = kCGPathFill;
		CGPathRelease(p);
		
		box = [[NSArray alloc] initWithObjects:f,t,b,l,r,nil];
	}
	return box;
}

- (NSArray*)glyphs
{
	if (!glyphs)
	{
		glyphs = [[NSMutableArray alloc] initWithCapacity:12];
		NSMutableArray* myGlyphs = (NSMutableArray*) glyphs;

		Text* text = [[Text new] autorelease];
		text.drawMode = kCGTextFillStroke;
		text.strokeWidth = 1.0;
		text.font = [UIFont systemFontOfSize:18];
		text.color = [UIColor blackColor];
		text.offset = CGPointMake(.5*cellSize.width-7, .5*cellSize.height-10);
		text.string = @"";
		
		// Numbers
		for (int i = 0; i <= 8; i++)
		{
			NSMutableArray* a = [[[NSMutableArray alloc] initWithArray:self.box copyItems:YES] autorelease];
			
			if (i)
			{
				Text* t = [[text copy] autorelease];
				t.string = [NSString stringWithFormat:@"%d",i];
				[a addObject:t];
			}
			
			[myGlyphs addObject:a];
		}
		
		// Mine
		{
			NSMutableArray* a = [[[NSMutableArray alloc] initWithArray:self.box copyItems:YES] autorelease];
			[[a objectAtIndex:0] setColor:[UIColor redColor]];
			
			Text* t = [[text copy] autorelease];
			t.string = @"X";
			[a addObject:t];
			
			[myGlyphs addObject:a];
		}
		
		// Hidden
		{
			NSMutableArray* a = [[[NSMutableArray alloc] initWithArray:self.box copyItems:YES] autorelease];
			[[a objectAtIndex:1] setColor:[[self.box objectAtIndex:2] color]];
			[[a objectAtIndex:2] setColor:[[self.box objectAtIndex:1] color]];
			[[a objectAtIndex:3] setColor:[[self.box objectAtIndex:4] color]];
			[[a objectAtIndex:4] setColor:[[self.box objectAtIndex:3] color]];
			
			[myGlyphs addObject:a];
		}
		
		// Flagged
		{
			NSMutableArray* a = [[[NSMutableArray alloc] initWithArray:[myGlyphs objectAtIndex:10] copyItems:YES] autorelease];

			Path* t = [[Path new] autorelease];
			CGMutablePathRef p = CGPathCreateMutable();
			CGPathAddRect(p, NULL, CGRectMake(0, 0, cellSize.width, cellSize.height));
			t.path = p;
			t.color = [UIColor colorWithRed:1.0 green:0.25 blue:0.0 alpha:0.35];
			t.drawMode = kCGPathFill;
			CGPathRelease(p);
			[a addObject:t];

			[myGlyphs addObject:a];
		}
	}
	return glyphs;
}

- (NSArray*)contents
{
	if (!contents)
	{
		contents = [[NSMutableArray alloc] initWithCapacity:[self.glyphs count]];
		NSMutableArray* myContents = (NSMutableArray*) contents;
		
		for (NSArray* symbol in self.glyphs)
		{
			// Create and get a pointer to context
			if (UIGraphicsBeginImageContextWithOptions != NULL)
			{
				UIGraphicsBeginImageContextWithOptions(cellSize, NO, 0.0);
			}
			else
			{
				UIGraphicsBeginImageContext(cellSize);
			}
			CGContextRef context = UIGraphicsGetCurrentContext();
			
			// Flop text rightside-up
			CGContextSetTextMatrix(context, CGAffineTransformMakeScale(1, -1));
			
			// Render box
			for (Glyph* g in symbol)
			{
				[g renderToContext:context];
			}
			
			// Create and store image
			[myContents addObject:UIGraphicsGetImageFromCurrentImageContext()];
			
			// Release image context, and continue
			UIGraphicsEndImageContext();
		}
	}
	return contents;
}


// The parent's designated initializer. Overriden to call the new DI.
- (id)initWithNibName:(NSString*)nibNameOrNil bundle:(NSBundle*)nibBundleOrNil
{
	return [self initWithNibName:nibNameOrNil bundle:nibBundleOrNil puzzle:nil];
}


// The new designated initializer.
- (id)initWithNibName:(NSString*)nibNameOrNil bundle:(NSBundle*)nibBundleOrNil puzzle:(ManagedPuzzle*)mp
{
    if (self = [super initWithNibName:nibNameOrNil bundle:nibBundleOrNil])
	{
		if (mp)
		{
			puzzle = [[Puzzle alloc] initWithManagedPuzzle:mp];
			managedPuzzle = [mp retain];
		}
		else
		{
			[self release];
			return nil;
		}
    }
    return self;
}


/*
// Implement loadView to create a view hierarchy programmatically, without using a nib.
- (void)loadView {
}
*/


// Implement viewDidLoad to do additional setup after loading the view, typically from a nib.
- (void)viewDidLoad
{
    [super viewDidLoad];
	self.navigationItem.titleView = self.titleLabel;
	
	[[NSNotificationCenter defaultCenter] addObserver:self
											 selector:@selector(unpause)
												 name:UIApplicationDidBecomeActiveNotification
											   object:nil];
	[[NSNotificationCenter defaultCenter] addObserver:self
											 selector:@selector(pause)
												 name:UIApplicationWillResignActiveNotification
											   object:nil];
	if (&UIApplicationDidEnterBackgroundNotification != NULL)
	{
		[[NSNotificationCenter defaultCenter] addObserver:self
												 selector:@selector(save)
													 name:UIApplicationDidEnterBackgroundNotification
												   object:nil];
	}
	
	[self setHeader];

	world = CGSizeMake(cellSize.width*puzzle.w, cellSize.height*puzzle.h);
	scale = 1.0;
	
	[self createTiles];
	tileBox = CGRectZero;
	
	[self sizeContent];
	CGPoint co = CGPointMake([managedPuzzle.scrollX floatValue], [managedPuzzle.scrollY floatValue]);
	if (CGPointEqualToPoint(co, self.scrollView.contentOffset))
	{
		[self scrollViewDidScroll:self.scrollView];
	}
	else
	{
		self.scrollView.contentOffset = co;
	}
	[self.scrollView flashScrollIndicators];
}


- (void)viewDidAppear:(BOOL)animated
{
	[super viewDidAppear:animated];
	
	UIApplication* app = [UIApplication sharedApplication];
	if ([app respondsToSelector:@selector(applicationState)] && (app.applicationState != UIApplicationStateActive))
	{
		[self pause];
	}
	else
	{
		[self startTimer];
	}
}


- (void)viewWillDisappear:(BOOL)animated
{
	[super viewWillDisappear:animated];
	
	[self stopTimer];
	[self save];
}


/*
// Override to allow orientations other than the default portrait orientation.
- (BOOL)shouldAutorotateToInterfaceOrientation:(UIInterfaceOrientation)interfaceOrientation {
    // Return YES for supported orientations
    return (interfaceOrientation == UIInterfaceOrientationPortrait);
}
*/

- (void)didReceiveMemoryWarning {
	// Releases the view if it doesn't have a superview.
    [super didReceiveMemoryWarning];
	
	// Release any cached data, images, etc that aren't in use.
}

- (void)viewDidUnload {
	// Release any retained subviews of the main view.
	// e.g. self.myOutlet = nil;
}


- (void)dealloc
{
	// Remove self as an observer
	[[NSNotificationCenter defaultCenter] removeObserver:self];
	
	// Clear weak delegate reference
	if (self.scrollView.delegate == self)
	{
		self.scrollView.delegate = nil;
	}

	[backButton release];
	[retryButton release];
	[daemonTimer invalidate];
	[daemonTimer release];
	[timerFormatter release];
	[box release];
	[glyphs release];
	[contents release];
	[puzzle release];
	[managedPuzzle release];
	[timerStart release];
	[self removeTiles];
	[titleLabel release];
	[toolbar release];
	[scrollView release];
	[pauseScreen release];
	[pauseLabel release];
	[modeControl release];
	[content release];
    [super dealloc];
}

#pragma mark Scroll view delegate methods

- (void)scrollViewDidScroll:(UIScrollView*)scrollView
{
	// Tile width and height (in view space)
	CGFloat w = tileSize.width;
	CGFloat h = tileSize.height;
	
	// Legal region (in view space)
	CGRect legal = CGRectMake(0, 0, world.width*scale, world.height*scale);
	
	// Displayed region (in view space)
	CGRect viewport = CGRectIntersection([self.content convertRect:self.scrollView.bounds fromView:self.scrollView], legal);
	
	// For existing tiles
	for (UIView* tv in [self.content subviews])
	{
		// If there's no overlap with viewport, toss the tile
		if (!CGRectIntersectsRect(viewport, tv.frame))
		{
			[tv removeFromSuperview];
			[extraTiles addObject:tv];
		}
	}
	
	NSInteger oldMinCol = tileBox.origin.x;
	NSInteger oldMinRow = tileBox.origin.y;
	NSInteger oldMaxCol = CGRectGetMaxX(tileBox)-1;
	NSInteger oldMaxRow = CGRectGetMaxY(tileBox)-1;
	
	// For new tiles
	tileBox = CGRectIntegral(CGRectApplyAffineTransform(viewport, CGAffineTransformMakeScale(1/w, 1/h)));
	for (int bw = tileBox.size.width, n = bw*tileBox.size.height-1; n >= 0; n--)
	{
		NSInteger c = tileBox.origin.x + n % bw;
		NSInteger r = tileBox.origin.y + n / bw;
		
		//	If missing, create
		if ((c < oldMinCol) || (c > oldMaxCol) || (r < oldMinRow) || (r > oldMaxRow))
		{
			UIView* tv = [self addTileForFrame:CGRectIntersection(CGRectMake(c*w, r*h, w, h), legal)];
			tv.layer.contents = (id) [[self imageForCell:CGPointMake(c, r)] CGImage];
			tv.tag = TAG_BASE + puzzle.w*r + c;
		}
	}
}

#pragma mark Tap detecting view delegate methods

- (void)tapDetectingView:(TapDetectingView*)tdView gotSingleTapAtPoint:(CGPoint)tapPoint
{
	if (![self cellForPoint:&tapPoint]) return;

	NSData* updated = nil;
	if (self.modeControl.selectedSegmentIndex == 0)
	{
		updated = [puzzle revealCell:tapPoint withCellBox:tileBox];
		if (puzzle.status != PuzzleStatusUnsolved)
		{
			[self stopTimer];
		}
	}
	else
	{
		updated = [puzzle toggleFlagOnCell:tapPoint withCellBox:tileBox];
	}

	CGPoint* points = (CGPoint*) [updated bytes];
	for (NSInteger i = [updated length]/sizeof(CGPoint) - 1; i >= 0; i--)
	{
		tapPoint = points[i];

		UIView* tv = [content viewWithTag:(TAG_BASE + puzzle.w*tapPoint.y + tapPoint.x)];
		tv.layer.contents = (id) [[self imageForCell:tapPoint] CGImage];
	}
	
	[self setHeader];
}


- (void)tapDetectingView:(TapDetectingView*)tdView gotDoubleTapAtPoint:(CGPoint)tapPoint
{
	[self tapDetectingView:tdView gotSingleTapAtPoint:tapPoint];
}

#pragma mark Extension methods

- (void)pause
{
	[self stopTimer];
	[self.view addSubview:self.pauseScreen];
	self.pauseLabel.text = self.titleLabel.text;
	
	self.pauseScreen.alpha = 0.0;
	[UIView beginAnimations:nil context:nil];
	self.pauseScreen.alpha = 1.0;
	[UIView commitAnimations];
}


- (void)unpause
{
	[UIView beginAnimations:nil context:nil];
	self.pauseScreen.alpha = 0.0;
	[UIView setAnimationDelegate:self];
	[UIView setAnimationDidStopSelector:@selector(finishUnpause:finished:context:)];
	[UIView commitAnimations];
}


- (void)finishUnpause:(NSString*)animationID finished:(NSNumber*)finished context:(void*)context
{
	[self startTimer];
	[self.pauseScreen removeFromSuperview];
}


- (void)save
{
	[puzzle saveToManagedPuzzle:managedPuzzle];
	managedPuzzle.scrollX = [NSNumber numberWithFloat:scrollView.contentOffset.x];
	managedPuzzle.scrollY = [NSNumber numberWithFloat:scrollView.contentOffset.y];
	
	NSError* error;
	if (![managedPuzzle.managedObjectContext save:&error])
	{
		NSLog(@"Unresolved error %@, %@", error, [error userInfo]);
	}
}


- (void)retry
{
	NSArray* vcs = self.navigationController.viewControllers;
	if ([vcs count] < 2) return;
	
	// A little hacky, but so what?
	RootViewController* root = (RootViewController*) [vcs objectAtIndex:[vcs count]-2];

	[self.navigationController popViewControllerAnimated:NO];
	[root addPuzzleWithWidth:puzzle.w height:puzzle.h mines:puzzle.nMines];
}


- (void)dismiss
{
	[self.navigationController popViewControllerAnimated:YES];
}


- (BOOL)cellForPoint:(CGPoint*)point
{
	if (!point) return FALSE;
	
	CGPoint cell = CGPointMake(floorf(point->x/(cellSize.width*scale)),
							   floorf(point->y/(cellSize.height*scale)));
	
	if ([puzzle isLegalCell:cell])
	{
		*point = cell;
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}


- (UIImage*)imageForCell:(CGPoint)pos
{
	if ([puzzle isFlaggedCell:pos])
	{
		return [self.contents objectAtIndex:11];
	}
	else if (![puzzle isRevealedCell:pos])
	{
		return [self.contents objectAtIndex:10];
	}
	else if ([puzzle isMinedCell:pos])
	{
		return [self.contents objectAtIndex:9];
	}
	else
	{
		return [self.contents objectAtIndex:[puzzle neighborsOfCell:pos]];
	}
}


- (void)setHeader
{
	NSTimeInterval et = puzzle.elapsedTime - [self.timerStart timeIntervalSinceNow];
	NSString* s = [self.timerFormatter stringFromDate:[NSDate dateWithTimeIntervalSinceReferenceDate:et]];
	
	titleLabel.text = [NSString stringWithFormat:@"%@ %2d/%2d",s,puzzle.nFlagged,puzzle.nMines];
	
	if (puzzle.status == PuzzleStatusFailed)
	{
		titleLabel.textColor = [UIColor colorWithRed:1.0 green:.20 blue:.20 alpha:1.0];
		modeControl.tintColor = [UIColor redColor];
		toolbar.tintColor = [UIColor redColor];
		
		self.navigationItem.rightBarButtonItem = self.retryButton;
		self.navigationItem.leftBarButtonItem = self.backButton;
	}
	else if (puzzle.status == PuzzleStatusSolved)
	{
		titleLabel.textColor = [UIColor colorWithRed:.20 green:1.0 blue:.20 alpha:1.0];
		modeControl.tintColor = [UIColor greenColor];
		toolbar.tintColor = [UIColor greenColor];

		self.navigationItem.leftBarButtonItem = self.backButton;
	}
}


- (void)startTimer
{
	// Under certain conditions, this function does nothing
	if (self.daemonTimer || (puzzle.status != PuzzleStatusUnsolved))
	{
		// I.e., if the timer's already running, or the puzzle's already solved ...
		return;
	}
	
	// We're probably getting too cute here, but:  We want to update the header at 1Hz.
	// We want the header to change at a regular rate, so we want to be pretty sure that
	// the ET will advance by 1s every time the header updates - more specifically, that
	// floor(ET) will increase by 1s every time the header updates.
	//
	// In general, therefore, we want the timer to fire when the ET is around XX.5s.
	// That way, even if the timer drifts from its 1Hz clock, there will very likely be
	// an effective 1s increase in ET every time it fires.
	//
	// Timers are scheduled relative to their original firing time, so we want to start
	// our repeating timer on a half-second ET.  This function makes that happen.

	// Record the start of playtime
	self.timerStart = [NSDate date];

	// Current ET
	NSTimeInterval et = puzzle.elapsedTime - [self.timerStart timeIntervalSinceNow];
	
	// Next 1/2-second ET
	NSTimeInterval next = floor(et-.5)+1.5;
	
	// Prepare and start a timer
	NSTimer* timer = [[[NSTimer alloc] initWithFireDate:[NSDate dateWithTimeIntervalSinceNow:next-et]
											   interval:1
												 target:self
											   selector:@selector(setHeader)
											   userInfo:nil
												repeats:YES] autorelease];
	[[NSRunLoop currentRunLoop] addTimer:timer forMode:NSDefaultRunLoopMode];
	
	// Store the timer (first canceling any existing one)
	[self.daemonTimer invalidate];
	self.daemonTimer = timer;
	
	// Perform the first header update immediately
	[self setHeader];
}


- (void)stopTimer
{
	// Do final update to puzzle's ET
	[puzzle addPlaytime:-[self.timerStart timeIntervalSinceNow]];
	
	// End playtime
	self.timerStart = nil;
	
	// Stop system timer
	[self.daemonTimer invalidate];
	self.daemonTimer = nil;
}


- (void)removeTiles
{
	for (UIView* tv in tiles)
	{
		// Remove tiles from content view
		[tv removeFromSuperview];
	}
	[tiles release];
	tiles = nil;
	
	[extraTiles release];
	extraTiles = nil;
}


- (void)createTiles
{
	if (tiles) [self removeTiles];
	
	tiles = [[NSMutableArray alloc] initWithCapacity:4];
	extraTiles = [[NSMutableSet alloc] initWithCapacity:4];
}


- (UIView*)addTileForFrame:(CGRect)frame
{
	UIView* tv = [extraTiles anyObject];
	if (tv)
	{
		// tv also retained by tiles
		[extraTiles removeObject:tv];
	}
	else
	{
		tv = [[UIView new] autorelease];
		tv.clearsContextBeforeDrawing = NO;
		tv.backgroundColor = [UIColor grayColor];
		[tiles addObject:tv];
	}
	
	[self.content addSubview:tv];
	tv.frame = frame;
	[tv setNeedsDisplay];
	return tv;
}


- (void)sizeContent
{
	self.content.frame = CGRectMake(0, 0, world.width*scale, world.height*scale);
	self.scrollView.contentSize = CGSizeMake(CGRectGetMaxX(self.content.frame),
											 CGRectGetMaxY(self.content.frame));
}


- (void)reload
{
	// Recycle all tiles
	for (UIView* tv in [self.content subviews])
	{
		[tv removeFromSuperview];
		[extraTiles addObject:tv];
	}
	tileBox = CGRectZero;
	
	// Trigger re-tiling
	[self scrollViewDidScroll:self.scrollView];
}

@end
