//
//  Puzzle.m
//  Demine
//
//  Created by Michael Heyeck on 4/9/10.
//  Copyright 2010 Fair Oaks Labs, Inc.. All rights reserved.
//

#import "Puzzle.h"


const static NSUInteger MAX_EDGE_SIZE	= 500;
const static char BIT_REVEALED			= (1<<6);
const static char BIT_FLAGGED			= (1<<5);
const static char BIT_MINED				= (1<<4);
const static char NEIGHBOR_MASK			= 0xf;


static inline BOOL is_legal(int w, int h, int x, int y)
{
	return (x>=0)&&(x<w)&&(y>=0)&&(y<h);
}

static inline char get_record(char* d, int w, int h, int x, int y)
{
	return is_legal(w, h, x, y)?d[w*y+x]:0;
}

static inline char is_mined(char* d, int w, int h, int x, int y)
{
	return (get_record(d, w, h, x, y)&BIT_MINED)?1:0;
}


@interface Puzzle ()

@property (nonatomic, assign, readonly) char* cells;

- (NSData*)revealCascadeAt:(CGPoint)pos withCellBox:(CGRect)cellBox;

@end


@implementation Puzzle

@synthesize w;
@synthesize h;
@synthesize nMines;
@synthesize nUncovered;
@synthesize nFlagged;
@synthesize elapsedTime;
@synthesize status;

- (char*)cells
{
	if (!cells)
	{
		cells = (char*) malloc(w*h);
	}
	return cells;
}


- (id)init
{
	return [self initWithWidth:0 height:0 mines:0];
}


- (id)initWithWidth:(NSUInteger)iw height:(NSUInteger)ih mines:(NSUInteger)im
{
	if (self = [super init])
	{
		w = MIN(MAX(iw, 2), MAX_EDGE_SIZE);
		h = MIN(MAX(ih, 2), MAX_EDGE_SIZE);
		
		nMines = MIN(im, (w-1)*(h-1));
	}
	return self;
}


- (id)initWithManagedPuzzle:(ManagedPuzzle*)puzzle
{
	if (self = [self initWithWidth:[puzzle.w integerValue] height:[puzzle.h integerValue] mines:[puzzle.nMines integerValue]])
	{
		nUncovered	= [puzzle.nUncovered integerValue];
		nFlagged	= [puzzle.nFlagged integerValue];
		elapsedTime	= [puzzle.elapsedTime doubleValue];
		status		= [puzzle.status integerValue];
		
		memcpy(self.cells, [puzzle.cells bytes], MIN([puzzle.cells length], w*h));
	}
	return self;
}


- (void)saveToManagedPuzzle:(ManagedPuzzle*)puzzle
{
	puzzle.w			= [NSNumber numberWithInteger:w];
	puzzle.h			= [NSNumber numberWithInteger:h];
	puzzle.nMines		= [NSNumber numberWithInteger:nMines];
	puzzle.nUncovered	= [NSNumber numberWithInteger:nUncovered];
	puzzle.nFlagged		= [NSNumber numberWithInteger:nFlagged];
	puzzle.elapsedTime	= [NSNumber numberWithDouble:elapsedTime];
	puzzle.status		= [NSNumber numberWithInteger:status];
	puzzle.cells		= [NSData dataWithBytes:self.cells length:w*h];
}


- (NSString*)description
{
	char symbols[2*w+1];
	memset(symbols, ' ', 2*w+1);
	symbols[0] = '\n';
	symbols[2*w] = '\0';
	NSMutableArray* a = [NSMutableArray arrayWithCapacity:h];

	char* d = self.cells;
	for (int y = 0; y < h; y++)
	{
		for (int x = 0; x < w; x++)
		{
			char cell = d[w*y+x];
			symbols[2*x+1] = (cell&BIT_MINED)?'X':('0'+(cell&NEIGHBOR_MASK));
		}
		[a addObject:[NSString stringWithCString:symbols encoding:NSASCIIStringEncoding]];
	}
	return [a componentsJoinedByString:@""];
}


- (void)generate
{
	char* d = self.cells;
	int size = w*h;

	// Reset counts and administrivia
	nUncovered	= 0;
	nFlagged	= 0;
	elapsedTime	= 0;
	status		= PuzzleStatusUnsolved;
	
	// To keep things reasonable, add or remove cells depending on density
	if (nMines > size/2)
	{
		// More than 50% mined: remove mines
		memset(d, BIT_MINED, size);
		for (int n=size; n>nMines;)
		{
			int idx = arc4random()%size;
			if (d[idx] & BIT_MINED)
			{
				n--;
				d[idx] &= ~BIT_MINED;
			}
		}
	}
	else
	{
		// Less than 50% mined: add mines
		memset(d, 0, size);
		for (int n=0; n<nMines;)
		{
			int idx = arc4random()%size;
			if (!(d[idx] & BIT_MINED))
			{
				n++;
				d[idx] |= BIT_MINED;
			}
		}
	}
	
	// Calculate neighbors
	for (int x = w-1; x >= 0; x--)
	{
		for (int y = h-1; y >= 0; y--)
		{
			d[w*y+x] |= (is_mined(d,w,h,x-1,y-1) +
						 is_mined(d,w,h,x+0,y-1) +
						 is_mined(d,w,h,x+1,y-1) +
						 is_mined(d,w,h,x-1,y+0) +
						 is_mined(d,w,h,x+1,y+0) +
						 is_mined(d,w,h,x-1,y+1) +
						 is_mined(d,w,h,x+0,y+1) +
						 is_mined(d,w,h,x+1,y+1) );
		}
	}
}


- (NSData*)revealCell:(CGPoint)pos withCellBox:(CGRect)cellBox
{
	NSMutableData* rv = [NSMutableData dataWithBytes:&pos length:sizeof(CGPoint)];
	
	int x = (int) pos.x;
	int y = (int) pos.y;
	int n = w*y+x;
	
	if ((status != PuzzleStatusUnsolved) || (!is_legal(w, h, x, y)) || (self.cells[n] & (BIT_REVEALED|BIT_FLAGGED)))
	{
		return nil;
	}

	self.cells[n] |= BIT_REVEALED;
	nUncovered++;

	if (self.cells[n] & BIT_MINED)
	{
		status = PuzzleStatusFailed;
	}
	else 
	{
		if ((self.cells[n]&NEIGHBOR_MASK) == 0)
		{
			[rv appendData:[self revealCascadeAt:pos withCellBox:cellBox]];
		}
		if ((nMines+nUncovered) == (w*h))
		{
			status = PuzzleStatusSolved;
		}
	}
	
	return rv;
}


- (NSData*)toggleFlagOnCell:(CGPoint)pos withCellBox:(CGRect)cellBox
{
	int x = (int) pos.x;
	int y = (int) pos.y;
	int n = w*y+x;
	
	if ((status != PuzzleStatusUnsolved) || (!is_legal(w, h, x, y)) || (self.cells[n] & BIT_REVEALED))
	{
		return nil;
	}
	
	if (self.cells[n] & BIT_FLAGGED)
	{
		self.cells[n] &= ~BIT_FLAGGED;
		nFlagged--;
	}
	else
	{
		self.cells[n] |= BIT_FLAGGED;
		nFlagged++;
	}
	
	return [NSData dataWithBytes:&pos length:sizeof(CGPoint)];
}


- (void)addPlaytime:(NSTimeInterval)dt
{
	elapsedTime += dt;
}


- (BOOL)isLegalCell:(CGPoint)pos
{
	return is_legal(w, h, pos.x, pos.y);
}


- (BOOL)isRevealedCell:(CGPoint)pos
{
	return (get_record(self.cells, w, h, pos.x, pos.y)&BIT_REVEALED)?1:0;
}


- (BOOL)isFlaggedCell:(CGPoint)pos
{
	return (get_record(self.cells, w, h, pos.x, pos.y)&BIT_FLAGGED)?1:0;
}


- (BOOL)isMinedCell:(CGPoint)pos
{
	return is_mined(self.cells, w, h, pos.x, pos.y);
}


- (NSUInteger)neighborsOfCell:(CGPoint)pos
{
	return get_record(self.cells, w, h, pos.x, pos.y)&NEIGHBOR_MASK;
}


- (void)dealloc
{
	if (cells) free(cells);
    [super dealloc];
}

#pragma mark Extension methods

- (NSData*)revealCascadeAt:(CGPoint)pos withCellBox:(CGRect)cellBox
{
	NSMutableData* rv = [NSMutableData data];
	
	unsigned int myW = w;
	unsigned int myH = h;
	
	int myUncovered = nUncovered;
	char* myCells = self.cells;
	
	unsigned int minX = cellBox.origin.x;
	unsigned int minY = cellBox.origin.y;
	unsigned int maxX = cellBox.origin.x + cellBox.size.width;
	unsigned int maxY = cellBox.origin.y + cellBox.size.height;
	
	NSInteger ql = 1;
	unsigned int* q = (unsigned int*) malloc(w*h*sizeof(unsigned int));
	q[0] = (((unsigned int) pos.y) << 16) | ((unsigned int) pos.x);
	unsigned int offsets[8][2] = {-1, -1, -1, +0, -1, +1, +0, -1, +0, +1, +1, -1, +1, +0, +1, +1};

	for (NSInteger i = 0; i < ql; i++)
	{
		unsigned int ox = q[i]&0xffff;
		unsigned int oy = q[i]>>16;
		
		for (int j = 0; j < 8; j++)
		{
			unsigned int x = ox + offsets[j][1];
			unsigned int y = oy + offsets[j][0];
			unsigned int n = myW*y+x;
			
			if ((x >= myW) || (y >= myH) || (myCells[n] & (BIT_REVEALED|BIT_FLAGGED)))
			{
				continue;
			}
			
			myCells[n] |= BIT_REVEALED;
			myUncovered++;
			
			if ((x >= minX) && (x < maxX) && (y >= minY) && (y < maxY))
			{
				pos = CGPointMake(x, y);
				[rv appendBytes:&pos length:sizeof(CGPoint)];
			}
			
			if ((myCells[n]&NEIGHBOR_MASK) == 0)
			{
				q[ql++] = (y<<16)|x;
			}
		}
	}
	
	free(q);
	nUncovered = myUncovered;
	return rv;
}

@end
